# $Id: Main.pm,v 1.14 2004/03/25 15:09:28 brain Exp $

use Socket;

our $KILLED;
our $CONNECTS;
our $START_TIME;
our $curdir;

our @provides = ();

sub depends
{
	print "Checking depends: @_\n" if $debug;
	foreach my $t (@_) {
		if ($t ne '') {
			my $satisfy = 0;
			foreach my $token (@provides) {
				if ($token eq $t) {
					$satisfy = 1;
				}
			}
			if (!$satisfy) {
				return 0;
			}
		}
	}
	return 1;
}

sub provides
{
	print "Adding provides: @_\n" if $debug;
	foreach my $t (@_) {
		if ($t ne '') {
			if (depends($t)) {
				print "\n\nERROR! Two modules are providing the same features (token: $t)\n";
				print "You probably have the same module loaded twice, or a feature conflict!\n";
				print "Bailing...\n";
				exit(0);
			}
			push @provides,$t;
		}
	}
}

sub general_init
{
	if ($0 =~ /(^\/.*?)[^\/]*$/){
		# if called with fullpath or if script is in a ENV{PATH}-directory
		$curdir=$1;
	}
	else {
		$0 =~ /(.*?)[^\/]*$/;
		$curdir=$ENV{PWD}."/".$1;
	}
	print "IRC Defender - Programmed by C.J.Edwards (Brain) - irc.chatspike.net\n\n";
	$START_TIME = time;
}

sub load_config
{
	print "Loading configuration file...\n";
	push @rehash_data, "Loading configuration file...";

        @provides = ();
        provides("core-v1");

	$cfg = "$curdir/defender.conf";
	$dir = $curdir;
	open (CONFIG, $cfg) or die "Can't locate config file! ($cfg): $!";
	while (chomp($pair=<CONFIG>))
	{
	        ( $var, $val ) = split ( "=", $pair );  # Split name/value pairs
	        $val =~ s/%(..)/pack("c",hex($1))/ge;   # Replace %nn with char
	        $dataValues{"$var"} = "$val";
	}
	close(CONFIG);
	$domain = $dataValues{"domain"};
	$nspass = $dataValues{"nickserv"};
	$dir = $dataValues{"datadir"};
	$killurl = $dataValues{"url"};
	$killmail = $dataValues{"mail"};
	$server = $dataValues{"server"};
	$server_re = $dataValues{"servregexp"};
	$botnick = $dataValues{"botnick"};
	$mychan = $dataValues{"channel"};
	$quitmsg = $dataValues{"quitmsg"};
	$port = $dataValues{"port"};
	$paranoia = $dataValues{"paranoia"};
	$botname = $dataValues{"fullname"};
	$oname = $dataValues{"opername"};
	$opass = $dataValues{"operpass"};
	$password = $dataValues{"password"};
	$netname    = $dataValues{"networkname"};
	$servername = $dataValues{"servername"};
	$numeric    = $dataValues{"numeric"};
	$serverdesc = $dataValues{"serverdesc"};
	$linkmodule = $dataValues{"linktype"};
	$logger = $dataValues{"logto"};
	$supportchannel = $dataValues{"supportchannel"};
}


sub reinit_modules
{
	&link_init;

	my @removed;
	my @added;
	my @unchanged;

	$modules = $dataValues{"modules"};
	our @modlist = split(",",$modules);
	
	foreach my $mod (@modlist) {
		my $added = 1;
		foreach my $old (@oldmods) {
			if ($old eq $mod) {
				$added = 0;
			}
		}
		if ($added == 1) {
			push @added, $mod;
		}
	}	

	foreach my $mod (@oldmods) {
		my $removed = 1;
		foreach my $new (@modlist) {
			if ($new eq $mod) {
				$removed = 0;
			}
		}
		if ($removed == 1) {
			push @removed, $mod;
		}
	}

	foreach my $mod (@modlist) {
		my $unchanged = 1;
		foreach my $n (@removed) {
			if ($mod eq $n) {
				$unchanged = 0;
			}
		}
		foreach my $n (@added) {
			if ($mod eq $n) {
				$unchanged = 0;
			}
		}
		if ($unchanged == 1) {
			push @unchanged, $mod;
		}
	}

	print "Rehashing.\nAdded: @added Removed: @removed Unchanged: @unchanged\n\n";

	foreach $mod (@removed) {
		print "Unloading: Modules/Scan/$mod.pm... ";
		push @rehash_data, "Unloading: Modules/Scan/$mod.pm... ";
		my $func = "no Modules::Scan::" . $mod;
		eval $func;
		print $@ if $@;
		if ($@) { push @rehash_data, $@; }
		print "OK!\n";
	}

	foreach $mod (@added) {
		print ("Loading: Modules/Scan/$mod.pm... ");
		push @rehash_data, "Loading: Modules/Scan/$mod.pm... ";
		require "Modules/Scan/$mod.pm";
		my $func = "Modules::Scan::" . $mod . "::init";
		eval $func;
		print $@ if $@;
		if ($@) { push @rehash_data, $@; }
		print "OK!\n";
	}

	foreach $mod (@unchanged) {
		print ("Re-initializing: Modules/Scan/$mod.pm... ");
		push @rehash_data, "Re-initializing: Modules/Scan/$mod.pm... ";
		my $func = "Modules::Scan::" . $mod . "::init";
		eval $func;
		print $@ if $@;
		if ($@) { push @rehash_data, $@; }
		print "OK!\n";
	}	
}

sub init_modules
{
	require "Modules/Link/$linkmodule.pm";
	print ("Using $linkmodule connection module (loaded OK)...\n");
	&link_init;

	print ("\nLoading modules...\n");

	$modules = $dataValues{"modules"};
	our @modlist = split(",",$modules);
	foreach $mod (@modlist) {
	        print ("Loading: Modules/Scan/$mod.pm... ");
	        require "Modules/Scan/$mod.pm";
		my $func = "Modules::Scan::" . $mod . "::init";
	        eval $func;
		print ("OK!\n");
	}

	require "Modules/Log/$logger.pm";
	if (!$debug)
	{
		print "Switching to $logger logging method from now\n";
		#my $startup = ("Modules::Log::" . $logger . "::init");
		eval "Modules::Log::$logger"->init();
		print $@ if $@;
	}
	else
	{
		print "Not enabling logging module because debugging is enabled\n";
		print "Will log to console.\n";
	}
}

sub rehash
{
	chdir $dir;
	our @rehash_data = ();
	our @oldmods = @modlist;
	&load_config;
	&reinit_modules;
	chdir '/';
}

sub check_params
{
	use Getopt::Long;
	GetOptions("debug" => \$debug, "help" => \$help);

	if (defined($help))
	{
               print "--debug                Don't daemonize, display all raw I/O on console\n";
               print "--help                 Display this help text\n";
               print "\nReleased under GPL by ChatSpike Development team (C) 2004\n";
               exit(0);
		
	}
	if (defined($debug))
	{
		$debug = 1;
	}
	else
	{
		$debug = 0;
	}
}

sub daemon
{
	use POSIX qw(setsid);

	if (!$debug)
	{
		chdir '/'                 or die "Can't chdir to /: $!";
		umask 0;
		open STDIN, '/dev/null'   or die "Can't read /dev/null: $!";
		open STDERR, '>/dev/null' or die "Can't write to /dev/null: $!";
		defined(my $pid = fork)   or die "Can't fork: $!";
		exit if $pid;
		setsid                    or die "Can't start a new session: $!";
	}

	$SIG{'INT'} = 'shutdown';
	$SIG{'HUP'} = 'rehash';
}

1;
